
############################################################################
# ========================================================================
# Copyright 2022 VMware, Inc.  All rights reserved. VMware Confidential
# ========================================================================
###
"""
This script allows the user to communicate with Infoblox DNS provider.

Required Functions
------------------
1. TestLogin: Function to verify provider credentials, used in the UI during DNS profile configuration.
2. GetAvailableDomains: Function to return available (discoverable) domains from the provider.
3. CreateOrUpdateDnsRecords: Function to create or update DNS records with the provider.
4. DeleteDnsRecords: Function to delete DNS records from the provider.

Required Exception Classes
--------------------------
1. CustomDnsAuthenticationErrorException: Raised when authentication fails.
2. CustomDnsRecordNotFoundException: Raised when given record not found, incase of update record request.
3. CustomDnsRecordAlreadyExistsException: Raised when the record already exists.
4. CustomDnsNotImplementedException: Raised when the functionality is not implemented.
5. CustomDnsGeneralException: Raised for other types of exceptions.

Helper Functions (Optional)
---------------------------
1. _verify_required_fields_in_auth_params: Internal function used to verify that all required auth parameters are provided.
2. _check_and_raise_auth_error: Internal function used to raise an exception if authentication fails.
3. _build_ipvxaddrs_objects: Internal function used to build ipv4addrs and ipv6addrs objects for payload data from a given list of ips.
4. _get_ips_by_host: Internal function used to return ips for a given record name.
5. _create_dns_record: Internal function used to create a DNS record with the provider.
6. _update_dns_record: Internal function used to update a given DNS record with the provider.
7. _delete_dns_record: Internal function used to delete a given DNS record from the provider.
"""

import requests
import json
import logging
import copy
from bs4 import BeautifulSoup


class CustomDnsAuthenticationErrorException(Exception):
    """
    Raised when authentication fails.
    """
    pass

class CustomDnsRecordNotFoundException(Exception):
    """
    Raised when given record not found, incase of update record request.
    """
    pass

class CustomDnsRecordAlreadyExistsException(Exception):
    """
    Raised when the record already exists.
    """
    pass


class CustomDnsNotImplementedException(Exception):
    """
    Raised when the functionality is not implemented.
    """
    pass

class CustomDnsGeneralException(Exception):
    """
    Raised for other types of exceptions.
    """
    pass


def _check_and_raise_auth_error(response):
    """
    Function to check authentication error for a given response and raise an exception for the same.
    """
    if response.status_code in [401, 403]:
        text = BeautifulSoup(response.text, 'html.parser').text
        raise CustomDnsAuthenticationErrorException(response.reason, response.status_code, response.url, text)


def _verify_required_fields_in_auth_params(auth_params):
    """
    Function to verify that all required auth parameters are provided.
    """
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    missing_req_params = []
    if 'server' not in auth_params:
        missing_req_params.append('server')
    if 'username' not in auth_params:
        missing_req_params.append('username')
    if 'password' not in auth_params:
        missing_req_params.append('password')
    if 'wapi_version' not in auth_params:
        missing_req_params.append('wapi_version')
    if missing_req_params:
        missing_req_params = ', '.join(missing_req_params)
        logger.error("all credentials are not provided, missing parameters[%s]" % missing_req_params)
        raise CustomDnsGeneralException("all credentials are not provided, missing parameters[%s]" % missing_req_params)


def _build_ipvxaddrs_objects(ips):
    """
    Function to build ipv4addrs and ipv6addrs objects for payload data from a given list of ips.
    """
    ipv4addrs = []
    ipv6addrs = []
    if ips['v4_ips']:
        for v4_ip in ips['v4_ips']:
            ipv4addrs.append({'ipv4addr': v4_ip})
    if ips['v6_ips']:
        for v6_ip in ips['v6_ips']:
            ipv6addrs.append({'ipv6addr': v6_ip})
    return ipv4addrs, ipv6addrs


def _get_ips_by_host(auth_params, record_name, ip_type='V4_V6'):
    """
    Function to return ips for a given record name.
    """
    username = auth_params.get('username')
    password = auth_params.get('password')
    server = auth_params.get('server')
    wapi_version = auth_params.get('wapi_version')
    dns_view = auth_params.get('dns_view', 'default')
    logger = logging.getLogger(auth_params.get('logger_name', ''))

    rest_url = 'https://' + server + '/wapi/' + wapi_version + \
            '/record:host?name=' + record_name + '&view=' + dns_view
    ipaddrs = []
    try:
        r = requests.get(url=rest_url, auth=(username, password), verify=False)
        _check_and_raise_auth_error(r)
        logger.info("record_name[%s], GET req[%s] status_code[%s]" % (record_name, rest_url, r.status_code))
        r_json = r.json()
        err_msg = str(r.status_code)
        if r.status_code == 200:
            if len(r_json) > 0:
                if ip_type == 'V4_ONLY' or ip_type == 'V4_V6':
                    if 'ipv4addrs' in r_json[0] and len(r_json[0]['ipv4addrs']) > 0:
                        for ipv4addr in r_json[0]['ipv4addrs']:
                            ipaddrs.append(ipv4addr['ipv4addr'])
                if ip_type == 'V6_ONLY' or ip_type == 'V4_V6':
                    if 'ipv6addrs' in r_json[0] and len(r_json[0]['ipv6addrs']) > 0:
                        for ipv6addr in r_json[0]['ipv6addrs']:
                            ipaddrs.append(ipv6addr['ipv6addr'])
                return ipaddrs
            else:
                err_msg += ": No host records found!"
        else:
            err_msg += ' : '  + r_json['text'] if 'text' in r_json else ''
            raise CustomDnsGeneralException(err_msg)
    except CustomDnsAuthenticationErrorException as e:
        raise
    except Exception as e:
        logger.error("error retrieving ips for a host record[%s] reason[%s]" %(record_name, str(e)))
        raise CustomDnsGeneralException("error retrieving ips for a host record[%s] reason[%s]" %(record_name, str(e)))

    
def _create_dns_record(auth_params, record_name, ips):
    """
    Function to create a DNS record.
    """
    username = auth_params.get('username')
    password = auth_params.get('password')
    server = auth_params.get('server')
    wapi_version = auth_params.get('wapi_version')
    dns_view = auth_params.get('dns_view', 'default')
    
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    ipv4addrs, ipv6addrs = _build_ipvxaddrs_objects(ips)
    rest_url = 'https://' + server + '/wapi/' + wapi_version + '/record:host?_return_fields=ipv4addrs,ipv6addrs'
    payload = json.dumps({'name': record_name,'view': dns_view, 'ipv4addrs': ipv4addrs, 'ipv6addrs': ipv6addrs })
    
    try:
        r = requests.post(url=rest_url, auth=(username, password),
                    verify=False, data=payload)
        _check_and_raise_auth_error(r)
        logger.info("record_name[%s], POST req[%s %s] status_code[%s]" % (record_name, rest_url, payload, r.status_code))
        r_json = r.json()
        if r.status_code == 200 or r.status_code == 201:
            return
        elif r.text and 'already exists' in r_json['text']:
            host_ips = _get_ips_by_host(auth_params, record_name)
            logger.info("record_name[%s], record ips on infoblox[%s]" %(record_name, host_ips))
            if set(ips['v4_ips']).issubset(set(host_ips)) and set(ips['v6_ips']).issubset(set(host_ips)):
                logger.info("record name[%s] for ipv4addrs %s and ipv6addrs %s already exists."%
                            (record_name, ips['v4_ips'], ips['v6_ips']))
            else:
                raise CustomDnsRecordAlreadyExistsException(r_json['text'])
    except CustomDnsAuthenticationErrorException as e:
        raise
    except CustomDnsRecordAlreadyExistsException as e:
        raise CustomDnsRecordAlreadyExistsException("record[%s] reason[%s]" %(record_name, str(e)))
    except Exception as e:
        logger.error("Error creating dns record %s on Infoblox, req[%s %s] rsp[%s]" %
                        (record_name, rest_url, payload, str(e)))
        raise CustomDnsGeneralException("Error creating dns record %s on Infoblox, reason[%s]" %(record_name, str(e)))


def _update_dns_record(auth_params, record_name, ips):
    """
    Function to update a given DNS record.
    """
    username = auth_params.get('username')
    password = auth_params.get('password')
    server = auth_params.get('server')
    wapi_version = auth_params.get('wapi_version')
    dns_view = auth_params.get('dns_view', 'default')
    
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    ipv4addrs, ipv6addrs = _build_ipvxaddrs_objects(ips)
    payload = json.dumps({'name': record_name,'view': dns_view, 'ipv4addrs': ipv4addrs, 'ipv6addrs': ipv6addrs })

    # Get the reference of the dns record
    host_ref = None
    rest_url = 'https://' + server + '/wapi/' + wapi_version + '/record:host?name=' + record_name
    try:
        r = requests.get(url=rest_url, auth=(username, password), verify=False)
        _check_and_raise_auth_error(r)
        logger.info("record_name[%s], GET req[%s] status_code[%s]" % (record_name, rest_url, r.status_code))
        r_json = r.json()
        if r.status_code == 200:
            host_ref = r_json[0]['_ref'] if len(r_json) > 0 and r_json[0]['_ref'] else None
        else:
            err_msg = str(r.status_code) + (' : '  + r_json['text'] if 'text' in r_json else '')
            raise CustomDnsGeneralException(err_msg)
    except CustomDnsAuthenticationErrorException as e:
        raise
    except Exception as e:
        logger.error("Error retrieving the record[%s] reason[%s]" % (record_name, str(e)))
        raise CustomDnsGeneralException("Error retrieving the record[%s] reason[%s]" % (record_name, str(e)))
    
    # Raise an error message if the record not found
    if not host_ref:
        err_msg = "record[%s] not found!" % record_name
        logger.error(err_msg)
        raise CustomDnsRecordNotFoundException(err_msg)

    rest_url = 'https://' + server + '/wapi/' + wapi_version + '/' + host_ref + '?_return_fields=ipv4addrs,ipv6addrs'
    try:
        r = requests.put(url=rest_url, auth=(username, password),
                    verify=False, data=payload)
        _check_and_raise_auth_error(r)
        logger.info("record_name[%s], PUT req[%s %s] status_code[%s]" % (record_name, rest_url, payload, r.status_code))
        r_json = r.json()
        if r.status_code == 200 or r.status_code == 201:
            return
        elif 'text' in r_json:
            raise CustomDnsGeneralException(r_json['text'])
    except CustomDnsAuthenticationErrorException as e:
        raise
    except Exception as e:
        logger.error("Error updating dns record %s on Infoblox, req[%s %s] rsp[%s]" %
                        (record_name, rest_url, payload, str(e)))
        raise CustomDnsGeneralException("Error updating dns record %s on Infoblox, reason[%s]" %(record_name, str(e)))


def _delete_dns_record(auth_params, record_name):
    """
    Function to delete a given DNS record.
    """
    username = auth_params.get('username')
    password = auth_params.get('password')
    server = auth_params.get('server')
    wapi_version = auth_params.get('wapi_version')
    dns_view = auth_params.get('dns_view', 'default')
    
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    rest_url = 'https://' + server + '/wapi/' + wapi_version +'/record:host?name=' + record_name

    try:
        # Get the reference of the dns record
        r = requests.get(url=rest_url, auth=(username, password), verify=False)
        _check_and_raise_auth_error(r)
        logger.info("record_name[%s], GET req[%s] status_code[%s]" % (record_name, rest_url, r.status_code))
        r_json = r.json()
        host_ref = None
        err_msg = "record[%s] not found!" % record_name
        if r.status_code == 200:
            host_ref = r_json[0]['_ref'] if len(r_json) > 0 and r_json[0]['_ref'] else None
            if not host_ref:
                logger.info(err_msg)
                return
            
            # Delete the record
            rest_url = 'https://' + server + '/wapi/' + \
                wapi_version + '/' + host_ref
            r = requests.delete(url=rest_url, auth=(username, password), verify=False)
            _check_and_raise_auth_error(r)
            logger.info("record_name[%s], DELETE req[%s] status_code[%s]" % (record_name, rest_url, r.status_code))
            r_json = r.json()
            if r.status_code == 200:
                return
        if 'text' in r_json:
            err_msg = str(r.status_code) + BeautifulSoup(r.text, 'html.parser').text
        logger.error(err_msg)
        raise CustomDnsGeneralException(err_msg)
    except CustomDnsAuthenticationErrorException as e:
        raise
    except Exception as e:
        logger.error("Error deleting the record[%s] reason[%s]" % (record_name, str(e)))
        raise CustomDnsGeneralException("Error deleting the record[%s] reason[%s]" % (record_name, str(e)))


def TestLogin(auth_params):
    """
    Function to validate user credentials. This function is called from DNS profile 
    configuration UI page.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication. These are script parameters provided while 
            creating a Custom DNS profile.
            Eg: auth_params can have following keys
            server: Server ip address of the custom DNS provider
            username: self explanatory
            password: self explanatory 
            logger_name: logger name   
    Returns
    -------
        Return True on success    
    Raises
    ------
        CustomDnsNotImplementedException: if this function is not implemented.
        CustomDnsAuthenticationErrorException: if authentication fails.
    """
    
    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("Inside F[TestLogin] auth_params[%s]", tmp_auth_params)
    server = auth_params.get('server', None)
    username = auth_params.get('username', None)
    password = auth_params.get('password', None)
    wapi_version = auth_params.get('wapi_version', None)
  
    if not server or not username or not password or not wapi_version:
        raise CustomDnsGeneralException("F[TestLogin] all credentials are not provided")
    try:
        schema_url = 'https://' + server + \
            '/wapi/' + wapi_version + '/?_schema'
        auth = (username, password)
        r = requests.get(
            url=schema_url, auth=auth, verify=False, timeout=30)
        logger.info("F[TestLogin] req[%s] status_code[%s]" % (schema_url, r.status_code))
        if r.status_code == 200:
            return True
        _check_and_raise_auth_error(r)
    except CustomDnsAuthenticationErrorException as e:
        raise
    except Exception as e:
        raise CustomDnsGeneralException("F[TestLogin] login failed reason[%s]" % str(e))


def GetAvailableDomains(auth_params):
    """
    Function to retrieve domains from the provider.
    Called from the DNS profile configuration to populate usable domains on the UI.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
    Returns
    -------
        domain_list: (list of str)
    Raises
    ------
        CustomDnsNotImplementedException: if this function is not implemented.
        CustomDnsGeneralException: if the api request fails.
    """

    raise CustomDnsNotImplementedException("F[GetAvailableDomains] not implemented in the script!")
         

def DeleteDnsRecords(records_info, auth_params):
    """
    Function to delete DNS records from the provider.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        record_info: (dict of str: dict)
            Records information with following keys.
            ips (dict of str: list of str): Keys are 'v4_ips' and 'v6_ips', values are list of IPs for each.
            dns_info (dict of str: dict): Keys are FQDNs, value contains FQDN metadata such as TTL, record type, etc,.
    Returns
    -------
        Return true on success.
    Raises
    ------
        CustomDnsAuthenticationErrorException: if authentication fails.
        CustomDnsGeneralException: if api request fails for any other reason.
    """
    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("records_info %s, auth_params %s"%(records_info, tmp_auth_params))
    for fqdn in records_info['dns_info'].keys():
        _delete_dns_record(auth_params, fqdn)
    return True


def CreateOrUpdateDnsRecords(new_records_info, old_records_info, auth_params):
    """
    Function to create/update DNS records with the provider.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        new_record_info: (dict of str: dict)
            New records information with following keys.
            ips (dict of str: list of str): Keys are 'v4_ips' and 'v6_ips', values are list of IPs for each.
            dns_info (dict of str: dict): Keys are FQDNs, value contains FQDN metadata such as TTL, record type, etc,.
        old_record_info: (dict of str: dict)
            Old records information with following keys.
            ips (dict of str: list of str): Keys are 'v4_ips' and 'v6_ips', values are list of IPs for each.
            dns_info (dict of str: dict): Keys are FQDNs, value contains FQDN metadata such as TTL, record type, etc,.
    Returns
    -------
        Returns true on success.
    Raises
    ------
        CustomDnsAuthenticationErrorException: if authentication fails.
        CustomDnsRecordAlreadyExistsException: if the record already exists.
        CustomDnsRecordNotFoundException: if the record not found, in case of update record request.
        CustomDnsGeneralException: if api request fails for any other reason.
    """
    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("new_records_info[%s], old_records_info[%s], auth_params[%s]"%(
                new_records_info, old_records_info,  tmp_auth_params))
    
    # Compare new_records_info and old_records_info, and accordingly 
    # perform create/update/delete action on the records.
    
    # Step 1. If new_records_info is empty, then delete all records in the old_records_info.
    # Happens in errback case or in vsvip update case where all DNS info is removed.
    if not new_records_info:
        DeleteDnsRecords(old_records_info, auth_params)
        return True

    # Step 2. If old_records_info is empty, then create all records in the new_records_info.
    if not old_records_info:
        for fqdn in new_records_info['dns_info'].keys():
            _create_dns_record(auth_params, fqdn, new_records_info['ips'])
        return True
    
    if new_records_info == old_records_info:
        logger.info("Both new_records_info and old_records_info are same, ignoring this action")
        return 

    # Step 3. If both new_records_info and old_records_info are not empty, 
    # compare both and perform required actions.
    # Step 3a. Perform create record operation for newly added fqdns 
    # and update record operation for updated fqdns.
    new_records = set(new_records_info['dns_info'].keys())
    old_records = set(old_records_info['dns_info'].keys())
    added_fqdns = new_records.difference(old_records)
    deleted_fqdns = old_records.difference(new_records)
    updated_fqdns = []
    if new_records_info['ips'] != old_records_info['ips']:
        updated_fqdns = new_records.intersection(old_records)
    logger.info("added fqdns[%s], updated fqdns[%s], deleted fqdns[%s]"%(added_fqdns, updated_fqdns, deleted_fqdns))
    for fqdn in added_fqdns:
        _create_dns_record(auth_params, fqdn, new_records_info['ips'])
    for fqdn in updated_fqdns:
        _update_dns_record(auth_params, fqdn, new_records_info['ips'])

    # Step 3b. Check for deleted fqdns and perform delete record operation for deleted fqdns.
    deleted_fqdns_record_info = {}
    deleted_fqdns_record_info['dns_info'] = {}
    deleted_fqdns_record_info['ips'] = old_records_info['ips']
    for fqdn in deleted_fqdns:
        deleted_fqdns_record_info['dns_info'][fqdn] = old_records_info['dns_info'][fqdn]
    if deleted_fqdns_record_info:
        DeleteDnsRecords(deleted_fqdns_record_info, auth_params)
    return True 



############################################################################
# ========================================================================
# Copyright 2022 VMware, Inc.  All rights reserved. VMware Confidential
# ========================================================================
###

"""
This script allows the user to communicate with Infoblox IPAM provider.

Required Functions
------------------
1. TestLogin: Function to verify provider credentials, used in the UI during IPAM profile configuration.
2. GetAvailableNetworksAndSubnets: Function to return available networks/subnets from the provider.
3. GetIpamRecord: Function to return the info of the given IPAM record.
4. CreateIpamRecord: Function to create an IPAM record with the provider.
5. DeleteIpamRecord: Funtion to delete a given IPAM record from the provider.
6. UpdateIpamRecord: Function to update a given IPAM record.

Required Exception Classes
--------------------------
1. CustomIpamAuthenticationErrorException: Raised when authentication fails.
2. CustomIpamRecordNotFoundException: Raised when given record not found.
3. CustomIpamNoFreeIpException: Raised when no free IP available in the given subnets/networks.
4. CustomIpamNotImplementedException: Raised when the functionality is not implemented.
5. CustomIpamGeneralException: Raised for other types of exceptions.

Optional
--------
1. _verify_required_fields_in_auth_params: Internal function used to verify that all required auth parameters are provided.
2. _check_and_raise_auth_error: Internal function used to raise an exception if authentication fails.
3. _get_api_paginated: Internal function used to return paginated response for a given url.
"""


# Script supports all standard python libraries and modules that are part of the Avi controller.
import requests
import time
import logging
import copy
from itertools import product
from bs4 import BeautifulSoup

class CustomIpamAuthenticationErrorException(Exception):
    """
    Raised when authentication fails.
    """
    pass

class CustomIpamRecordNotFoundException(Exception):
    """
    Raised when given record not found.
    """
    pass

class CustomIpamNoFreeIpException(Exception):
    """
    Raised when no free IP available in the given subnets/networks.
    """
    pass

class CustomIpamNotImplementedException(Exception):
    """
    Raised when the functionality is not implemented.
    """
    pass

class CustomIpamGeneralException(Exception):
    """
    Raised for other types of exceptions.
    """
    pass


def _verify_required_fields_in_auth_params(auth_params):
    missing_req_params = []
    if 'server' not in auth_params:
        missing_req_params.append('server')
    if 'username' not in auth_params:
        missing_req_params.append('username')
    if 'password' not in auth_params:
        missing_req_params.append('password')
    if 'wapi_version' not in auth_params:
        missing_req_params.append('wapi_version')
    if missing_req_params:
        missing_req_params = ', '.join(missing_req_params)
        raise CustomIpamGeneralException("all credentials are not provided, missing parameters[%s]" % missing_req_params)


def _check_and_raise_auth_error(response):
    if response.status_code in [401, 403]:
        text = BeautifulSoup(response.text, 'html.parser').text
        raise CustomIpamAuthenticationErrorException(response.reason, response.status_code, response.url, text)


def _get_api_paginated(auth_params, dest_url, data_key, results_per_page, api_timeout, ip_type):

    """
    Function used to return paginated response for a given dest_url. 
    This function is called from GetAvailableNetworksAndSubnets function.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        dest_url: (str)
            URL for which pagination is desired.
        data_key: (str)
            Dictionary key to look for in the results.
        results_per_page: (str)
            self explanatory
        api_timeout : (str)
            self explanatory
        ip_type: (str)
            IP type supported by the network/subnet used in the dest_url. Allowed values: V4_ONLY, V6_ONLY.
    Returns
    -------
        data_vals: (list of dict of str : str)
            each dict has 5 keys: network, v4_subnet, v6_subnet, v4_available_ip_count, v6_available_ip_count
            v4_available_ip_count and v6_available_ip_count are optional, currenty this function returns the first 3 keys. returning counts is TBD.
    Raises
    ------
        CustomIpamGeneralException: if the given request fails.
    """

    logger = logging.getLogger(auth_params.get('logger_name', ''))
    data_vals = []
    network_view = auth_params.get('network_view', 'default')
    page_url = dest_url + '&_paging=1&_max_results=%d&_return_as_object=1' % results_per_page
    r = requests.get(
        url=page_url, auth=(auth_params['username'], auth_params['password']), verify=False, timeout=30)
    logger.info("F[_get_api_paginated] req[%s] status_code[%s]" % (page_url, r.status_code))
    _check_and_raise_auth_error(r)
    if r.status_code == 200:
        resp = r.json()
        for unit in resp['result']:
            if data_key not in unit:
                continue
            sub_value = unit[data_key]
            # Since infoblox doesn't return network values, dummy value can be given to nw_id 
            # from here or from UI during IPAM profile configuration.
            nw_id = 'infoblox--' + network_view + '--'
            v4_subnet = v6_subnet = None
            if ip_type == 'V4_ONLY':
                v4_subnet = sub_value
                nw_id += v4_subnet.split('/')[0] + '-' + str(v4_subnet.split('/')[1])
            elif ip_type == 'V6_ONLY':
                v6_subnet = sub_value
                nw_id += v6_subnet.split('/')[0] + '-' + str(v6_subnet.split('/')[1])
            data_vals.append({'network': nw_id, 'v4_subnet': v4_subnet, 'v6_subnet': v6_subnet})

        next_page_id = resp.get('next_page_id', '')
        page_start_time = time.time()
        elapsed_time = 0
        while next_page_id and elapsed_time < api_timeout:
            page_url = dest_url + '&_page_id=%s' % next_page_id
            r = requests.get(
                url=page_url, auth=(auth_params['username'], auth_params['password']), verify=False, timeout=30)
            logger.info("F[_get_api_paginated] req[%s] status_code[%s]" % (page_url, r.status_code))
            _check_and_raise_auth_error(r)
            if r.status_code != 200:
                break
            resp = r.json()
            for unit in resp['result']:
                if data_key not in unit:
                    continue
                sub_value = unit[data_key]
                nw_id = 'infoblox--' + network_view + '--'
                v4_subnet = v6_subnet = None
                if ip_type == 'V4_ONLY':
                    v4_subnet = sub_value
                    nw_id += v4_subnet.split('/')[0] + '-' + str(v4_subnet.split('/')[1])
                elif ip_type == 'V6_ONLY':
                    v6_subnet = sub_value
                    nw_id += v6_subnet.split('/')[0] + '-' + str(v6_subnet.split('/')[1])
                data_vals.append({'network': nw_id, 'v4_subnet': v4_subnet, 'v6_subnet': v6_subnet})

            next_page_id = resp.get('next_page_id', '')
            elapsed_time = int(time.time() - page_start_time)
        return data_vals
    err_msg = r.status_code + ' : '  + r.text
    logger.error("F[GetAvailableNetworksAndSubnets] req[%s] ip_type[%s] err[%s]" % (page_url, ip_type, err_msg))
    raise CustomIpamGeneralException("F[GetAvailableNetworksAndSubnets] req[%s] ip_type[%s] err[%s]" % (page_url, ip_type, err_msg))


def GetIpamRecord(auth_params, record_info):
    """
    Function to return the info of the given IPAM record.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        record_info: (dict of str: str)
            id (str): uuid of vsvip.
            fqdns (list of str): list of fqdn from dns_info in vsvip. 
    Returns
    -------
        alloc_info(dict of str: str): 
            Dictionary of following keys
            v4_ip (str): IPv4 of the record
            v4_subnet (str): IPv4 subnet
            v6_ip (str): IPv6 of the record
            v6_subnet (str): IPv6 subnet
            network (str): network id/name
    Raises
    ------
        CustomIpamNotImplementedException: if this function is not implemented.
        CustomIpamGeneralException: if the api request fails.
    """

    raise CustomIpamNotImplementedException("F[GetIpamRecord] not implemented in the script!")


def TestLogin(auth_params):
    """
    Function to validate user credentials. This function is called from IPAM profile 
    configuration UI page.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication. These are script parameters provided while 
            creating a Custom IPAM profile.
            Eg: auth_params can have following keys
            server: Server ip address of the custom IPAM provider
            username: self explanatory
            password: self explanatory 
            logger_name: logger name   
    Returns
    -------
        Return True on success    
    Raises
    ------
        CustomIpamNotImplementedException: if this function is not implemented.
        CustomIpamAuthenticationErrorException: if authentication fails.
    """
    
    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("Inside F[TestLogin] auth_params[%s]", tmp_auth_params)
    server = auth_params.get('server', None)
    username = auth_params.get('username', None)
    password = auth_params.get('password', None)
    wapi_version = auth_params.get('wapi_version', None)
  
    if not server or not username or not password or not wapi_version:
        raise CustomIpamGeneralException("F[TestLogin] all credentials are not provided")
    try:
        schema_url = 'https://' + server + \
            '/wapi/' + wapi_version + '/?_schema'
        auth = (username, password)
        r = requests.get(
            url=schema_url, auth=auth, verify=False, timeout=30)
        logger.info("F[TestLogin] req[%s] status_code[%s]" % (schema_url, r.status_code))
        if r.status_code == 200:
            return True
        _check_and_raise_auth_error(r)
    except Exception as e:
        raise CustomIpamGeneralException("F[TestLogin] login failed reason[%s]" % str(e))
        

def GetAvailableNetworksAndSubnets(auth_params, ip_type):
    """ 
    Function to retrieve networks/subnets from the provider.
    Called from the IPAM profile configuration to populate usable subnets on the UI.
    Note: Subnets are represented in CIDR format.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        ip_type: (str)
            IP type supported by the networks/subnets. Allowed values: V4_ONLY, V6_ONLY and V4_V6.
    Returns
    -------
        subnet_list: (list of dict of str : str)
            network (str): network id/name
            v4_subnet (str): V4 subnet
            v6_subnet (str): V6 subnet
            v4_available_ip_count (str): V4 free ips count of the network/v4_subnet
            v6_available_ip_count (str): V6 free ips count of the network/v6_subnet
        each dict has 5 keys: network, v4_subnet, v6_subnet, v4_available_ip_count, v6_available_ip_count
        v4_available_ip_count and v6_available_ip_count are optional, currenty this function returns the first 3 keys. returning counts is TBD.
    Raises
    ------
        None
    """

    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("Inside F[GetAvailableNetworksAndSubnets] auth_params[%s] ip_type[%s]", tmp_auth_params, ip_type)
    network_view = auth_params.get('network_view', 'default')
    base_url = 'https://' + auth_params['server'] + '/wapi/' + auth_params['wapi_version']
    data_key = 'network'
    results_per_page = 1000
    api_timeout = 600

    subnet_list = []
    if ip_type in ['V4_ONLY', 'V4_V6']:
        rest_url = base_url + '/network?network_view=' + network_view
        v4_subnets = _get_api_paginated(auth_params, rest_url, data_key, results_per_page, api_timeout, 'V4_ONLY')
        subnet_list.extend(v4_subnets)
    if ip_type in ['V6_ONLY', 'V4_V6']:
        rest_url = base_url + '/ipv6network?network_view=' + network_view
        v6_subnets = _get_api_paginated(auth_params, rest_url, data_key, results_per_page, api_timeout, 'V6_ONLY')
        subnet_list.extend(v6_subnets)
    return subnet_list


def CreateIpamRecord(auth_params, record_info):
    """
    Implements a Custom Rest API to create an IPAM record.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        record_info: (dict of str: str)
            New record information with following keys.
            id (str): uuid of vsvip.
            fqdns (list of str): list of fqdn from dns_info in vsvip.
            preferred_ip (str): the vsvip IPv4 if it's already configured by the user.
            preferred_ip6 (str): the vsvip IPv6 if it's already configured by the user.
            allocation_type (str): IP allocation type. Allowed values: V4_ONLY, V6_ONLY and V4_V6.
            nw_and_subnet_list (list of tuples : str): List of networks and subnets to use for new IPAM
            record IP allocation. Each tuple has 3 values (network, v4_subnet, v6_subnet).
    Returns
    -------
        alloc_info(dict of str: str): 
            Dictionary of following keys
            v4_ip (str): allocated IPv4
            v4_subnet (str): subnet used for IPv4 allocation.
            v6_ip (str): allocated IPv6
            v6_subnet (str): subnet used for IPv6 allocation.
            network (str): network used for IPv4/IPv6 allocation.
    Raises
    ------
        CustomIpamNoFreeIpException: if no free ip available.
        CustomIpamGeneralException: if create record fails for any other reason.
    """

    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("Inside F[CreateIpamRecord] auth_params[%s] record_info[%s]", tmp_auth_params, record_info)
    network_view = auth_params.get('network_view', 'default')
    dns_view = auth_params.get('dns_view', 'default')
    id = record_info.get('id', None)
    fqdns = record_info.get('fqdns', None)
    allocation_type = record_info.get('allocation_type', None)
    nw_and_subnet_list = record_info.get('nw_and_subnet_list', [])
    preferred_ip = record_info.get('preferred_ip', None)
    preferred_ip6 = record_info.get('preferred_ip6', None)
    name = id
    
    base_uri = 'https://' + auth_params['server'] + '/wapi/' + auth_params['wapi_version']
    rest_url = base_uri + '/record:host?_return_fields=ipv4addrs,ipv6addrs'

    # Step 1: If the preferred_ip/preferred_ip6 is set, call specific rest URL according to the allocation_type.
    # <TODO>

    # Step 2: If the nw_and_subnet_list is empty call GetAvailableNetworksAndSubnets() to use any available
    # subnets or networks from the Custom IPAM provider.
    v4_subnet_list = v6_subnet_list = None
    if len(nw_and_subnet_list) == 0:
        if allocation_type == 'V4_V6':
            raise CustomIpamGeneralException("F[CreateIpamRecord] No V4+V6 networks found in Infoblox IPAM provider!")
        if allocation_type == 'V4_ONLY':
            nw_and_subnet_list = v4_subnet_list = GetAvailableNetworksAndSubnets(auth_params, 'V4_ONLY')
        if allocation_type == 'V6_ONLY':
            nw_and_subnet_list = v6_subnet_list = GetAvailableNetworksAndSubnets(auth_params, 'V6_ONLY')
        if len(nw_and_subnet_list) == 0:
            logger.error("F[CreateIpamRecord] No discovered networks and subnets found in Infoblox IPAM provider!")
            raise CustomIpamGeneralException("F[CreateIpamRecord] No discovered networks and subnets found in Infoblox IPAM provider!")

    # Step 3: Based on the allocation_type, build payload data and call specific rest URL.
    err_msgs = []
    if allocation_type == 'V4_V6':
        for sub in nw_and_subnet_list:
            v4_subnet = sub[1]
            v6_subnet = sub[2]
            if v4_subnet and v6_subnet:
                nw_str_ip4 = 'func:nextavailableip:' + v4_subnet + ',' + network_view
                nw_str_ip6 = 'func:nextavailableip:' + v6_subnet + ',' + network_view
                payload = '{"ipv4addrs": [{"ipv4addr": "' + nw_str_ip4 + '"}], "ipv6addrs": [{"ipv6addr": "' + nw_str_ip6 + '"}], "name": "' + \
                                name + '","configure_for_dns": false,"view": "' + dns_view + '"}'
            else:
                continue
            r = requests.post(url=rest_url, auth=(auth_params['username'], auth_params['password']),
                            verify=False, data=payload)
            logger.info("F[CreateIpamRecord] req[%s %s] status_code[%s]" % (rest_url, payload, r.status_code))
            _check_and_raise_auth_error(r)
            if r.status_code not in [200, 201]:
                emsg = "req[%s %s], rsp[%s, %s]" % (rest_url, payload, r.status_code, r.text)
                err_msgs.append(emsg)
                continue
            r_json = r.json()
            alloc_info = {}
            if 'ipv4addrs' in r_json and r_json['ipv4addrs'][0]['ipv4addr']:
                alloc_info['v4_ip'] = r_json['ipv4addrs'][0]['ipv4addr']
                alloc_info['v4_subnet'] = v4_subnet
            if 'ipv6addrs' in r_json and r_json['ipv6addrs'][0]['ipv6addr']:
                alloc_info['v6_ip'] = r_json['ipv6addrs'][0]['ipv6addr']
                alloc_info['v6_subnet'] = v6_subnet
            return alloc_info
        if len(err_msgs) == len(nw_and_subnet_list) and all('Cannot find 1 available IP address' in e for e in err_msgs):
            logger.error("F[CreateIpamRecord] No free available IP reason[%s]" % (err_msgs))
            raise CustomIpamNoFreeIpException("F[CreateIpamRecord] No free available IP reason[%s]" % (err_msgs))
        logger.error("F[CreateIpamRecord] Error creating host record[%s] reason[%s]" % (name, err_msgs))
        raise CustomIpamGeneralException("F[CreateIpamRecord] Error creating host record[%s] reason[%s]" % (name, err_msgs))

    else:
        for sub in nw_and_subnet_list:
            if v4_subnet_list or v6_subnet_list:
                v4_subnet = sub.get('v4_subnet','')
                v6_subnet = sub.get('v6_subnet','')
            else:
                v4_subnet = sub[1]
                v6_subnet = sub[2]
            if allocation_type in 'V4_ONLY' and v4_subnet:
                nw_str = 'func:nextavailableip:' + v4_subnet + ',' + network_view
                payload = '{"ipv4addrs": [{"ipv4addr": "' + nw_str + '"}],"name": "' + name + \
                        '","configure_for_dns": false,"view": "' + dns_view + '"}'
            elif allocation_type in 'V6_ONLY' and v6_subnet:
                nw_str = 'func:nextavailableip:' + v6_subnet + ',' + network_view
                payload = '{"ipv6addrs": [{"ipv6addr": "' + nw_str + '"}],"name": "' + name + \
                            '","configure_for_dns": false,"view": "' + dns_view + '"}'
            else:
                continue        
            r = requests.post(url=rest_url, auth=(auth_params['username'], auth_params['password']),
                            verify=False, data=payload)
            logger.info("F[CreateIpamRecord] req[%s %s] status_code[%s]" % (rest_url, payload, r.status_code))
            _check_and_raise_auth_error(r)
            if r.status_code not in [200, 201]:
                emsg = "req[%s %s], rsp[%s, %s]" % (rest_url, payload, r.status_code, r.text)
                err_msgs.append(emsg)
                continue
            r_json = r.json()           
            alloc_info = {}
            if 'ipv4addrs' in r_json and r_json['ipv4addrs'][0]['ipv4addr']:
                alloc_info['v4_ip'] = r_json['ipv4addrs'][0]['ipv4addr']
                alloc_info['v4_subnet'] = v4_subnet
            if 'ipv6addrs' in r_json and r_json['ipv6addrs'][0]['ipv6addr']:
                alloc_info['v6_ip'] = r_json['ipv6addrs'][0]['ipv6addr']
                alloc_info['v6_subnet'] = v6_subnet
            return alloc_info
        if len(err_msgs) == len(nw_and_subnet_list) and all('Cannot find 1 available IP address' in e for e in err_msgs):
            logger.error("F[CreateIpamRecord] No free available IP reason[%s]" % (err_msgs))
            raise CustomIpamNoFreeIpException("F[CreateIpamRecord] No free available IP reason[%s]" % (err_msgs))
        logger.error("F[CreateIpamRecord] Error creating host record[%s] reason[%s]" % (name, err_msgs))
        raise CustomIpamGeneralException("F[CreateIpamRecord] Error creating host record[%s] reason[%s]" % (name, err_msgs))
        

def DeleteIpamRecord(auth_params, record_info):
    """
    Implements a Custom Rest API to delete an IPAM record.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        record_info: (dict of str: str)
            Record to be deleted. Has following keys.
            id (str): uuid of vsvip.
            fqdns (list of str): list of fqdn from dns_info in vsvip.
    Returns
    -------
        True on successfully deleting the record.
    Raises
    ------
        CustomIpamRecordNotFoundException: if the given record not found
        CustomIpamGeneralException: if delete record request fails.
    """

    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("Inside F[DeleteIpamRecord] auth_params[%s] record_info[%s]", tmp_auth_params, record_info)
    id = record_info.get('id', None)
    fqdns = record_info.get('fqdns', None)
    name = id

    # Step 1: Get the reference of the given IPAM record
    rest_url = 'https://' + auth_params['server'] + '/wapi/' + auth_params['wapi_version'] + \
                '/record:host?name=' + name
    try:
        r = requests.get(url=rest_url, auth=(auth_params['username'], auth_params['password']), 
                            verify=False, timeout=30)
        logger.info("F[DeleteIpamRecord] req[%s] status_code[%s]" % (rest_url, r.status_code))
        _check_and_raise_auth_error(r)
        r_json = r.json()
        host_ref = None
        err_msg = "F[DeleteIpamRecord] Record[%s] not found!" % name
        if r.status_code == 200:
            host_ref = r_json[0]['_ref'] if len(r_json) > 0 and r_json[0]['_ref'] else None
            if not host_ref:
                logger.error(err_msg)
                raise CustomIpamRecordNotFoundException(err_msg)
            # Step 2: Delete the record
            rest_url = 'https://' + auth_params['server'] + '/wapi/' + \
                auth_params['wapi_version'] + '/' + host_ref
            r = requests.delete(url=rest_url, auth=(
                auth_params['username'], auth_params['password']), verify=False)
            logger.info("F[DeleteIpamRecord] req[%s] status_code[%s]" % (rest_url, r.status_code))
            _check_and_raise_auth_error(r)
            r_json = r.json()
            if r.status_code == 200:
                return True
        if 'text' in r_json:
            err_msg += r.status_code + " : " + BeautifulSoup(r.text, 'html.parser').text
        logger.error(err_msg)
        raise CustomIpamRecordNotFoundException(err_msg)
    except Exception as e:
        logger.error("F[DeleteIpamRecord] Error deleting the record[%s] reason[%s]" % (name, str(e)))
        raise CustomIpamGeneralException("F[DeleteIpamRecord] Error deleting the record[%s] reason[%s]" % (name, str(e)))


def UpdateIpamRecord(auth_params, new_record_info, old_record_info):
    """
    Function to handle update IPAM record requests. Eg: Change of allocation_type from 
    V4_ONLY to V6_ONLY.
    Args
    ----
        auth_params: (dict of str: str)
            Parameters required for authentication.
        new_record_info: (dict of str: str)
            New record information with following keys.
            id (str): uuid of vsvip.
            fqdns (list of str): list of fqdn from dns_info in vsvip.
            preferred_ip (str): the vsvip IPv4 if it's already configured by the user.
            preferred_ip6 (str): the vsvip IPv6 if it's already configured by the user.
            allocation_type (str): IP allocation type. Allowed values: V4_ONLY, V6_ONLY and V4_V6.
            nw_and_subnet_list (list of tuples : str): List of networks and subnets to use for an 
                IPAM record IP allocation. Each tuple has 3 values (network, v4_subnet, v6_subnet)
        old_record_info: (dict of str: str)
            Old record information with following keys.
            id (str): uuid of vsvip.
            fqdns (list of str): list of fqdn from dns_info in vsvip of an old record.
            preferred_ip (str): old record's preferred IPv4.
            preferred_ip6 (str): old record's preferred IPv6.
            allocation_type (str): old record's IP allocation type. Allowed values: V4_ONLY, V6_ONLY and V4_V6.
            nw_and_subnet_list (list of tuples : str): List of networks and subnets used for an old IPAM
                record IP allocation. Each tuple has 3 values (network, v4_subnet, v6_subnet)
    Returns
    -------
        alloc_info(dict of str: str): 
            Dictionary of following keys
            v4_ip (str): allocated IPv4
            v4_subnet (str): subnet used for IPv4 allocation.
            v6_ip (str): allocated IPv6
            v6_subnet (str): subnet used for IPv6 allocation.
            network (str): network used for IPv4/IPv6 allocation.
    Raises
    ------
        CustomIpamNotImplementedException: if this function or specific update requests not implemented.
        CustomIpamRecordNotFoundException: if the given record not found
        CustomIpamGeneralException: if the api request fails for any other reason.
    """

    _verify_required_fields_in_auth_params(auth_params)
    logger = logging.getLogger(auth_params.get('logger_name', ''))
    tmp_auth_params = copy.deepcopy(auth_params)
    tmp_auth_params['password'] = '<sensitive>'
    logger.info("Inside F[UpdateIpamRecord] auth_params[%s] new_record_info[%s] old_record_info[%s]", tmp_auth_params, new_record_info, old_record_info)
    id = new_record_info.get('id', None)
    new_fqdns = new_record_info.get('fqdns', None)
    old_fqdns = old_record_info.get('fqdns', None)
    preferred_ip = new_record_info.get('preferred_ip', None)
    preferred_ip6 = new_record_info.get('preferred_ip6', None)
    new_nw_and_subnet_list = new_record_info.get('nw_and_subnet_list', [])
    old_nw_and_subnet_list = old_record_info.get('nw_and_subnet_list', [])
    new_allocation_type = new_record_info.get('allocation_type', None)
    old_allocation_type = old_record_info.get('allocation_type', None)
    network_view = auth_params.get('network_view', 'default')
    
    # fqdn change: <TODO>
    # record update with preferred ipv4/ipv6: <TODO>
    # nw_and_subnet_list change: <TODO>
    # nw_and_subnet_list and allocation_type change: <TODO>

    # Note: Only allocation type change is handled here as an example update request. This is an in-place update.
    # Hence raise CustomIpamNotImplementedException exception for other update requests.
    if new_allocation_type == old_allocation_type or  not len(new_nw_and_subnet_list) == 1:
        logger.error("F[UpdateIpamRecord] script supports update allocation type only.")
        raise CustomIpamNotImplementedException("F[UpdateIpamRecord] script supports update allocation type only.")
    # network or subnet change along with allocation type update is also not supported.
    if old_allocation_type == 'V4_V6' or new_allocation_type == 'V4_V6':
        if (new_allocation_type == 'V4_ONLY' or old_allocation_type == 'V4_ONLY') and new_nw_and_subnet_list[0][1] != old_nw_and_subnet_list[0][1]:
            logger.error("F[UpdateIpamRecord] script supports update allocation type only.")
            raise CustomIpamNotImplementedException("F[UpdateIpamRecord] script supports update allocation type only, subnet change not supported!")
        if (new_allocation_type == 'V6_ONLY' or old_allocation_type == 'V6_ONLY') and new_nw_and_subnet_list[0][2] != old_nw_and_subnet_list[0][2]:
            logger.error("F[UpdateIpamRecord] script supports update allocation type only.")
            raise CustomIpamNotImplementedException("F[UpdateIpamRecord] script supports update allocation type only, subnet change not supported!")

    # Step 1: Get the reference of the given IPAM record
    name = id
    host_ref = None
    rest_url = 'https://' + auth_params['server'] + '/wapi/' + auth_params['wapi_version'] + \
                '/record:host?name=' + name
    try:
        r = requests.get(url=rest_url, auth=(auth_params['username'], auth_params['password']),
                            verify=False, timeout=30)
        logger.info("F[UpdateIpamRecord] req[%s] status_code[%s]" % (rest_url, r.status_code))
        _check_and_raise_auth_error(r)
        r_json = r.json()
        if r.status_code == 200:
            host_ref = r_json[0]['_ref'] if len(r_json) > 0 and r_json[0]['_ref'] else None
        else:
            err_msg = r.status_code 
            err_msg += ' : '  + r_json['text'] if 'text' in r_json else None
            logger.error("F[UpdateIpamRecord] Error retrieving the record[%s] reason[%s]" % (name, err_msg))
            raise CustomIpamRecordNotFoundException("F[UpdateIpamRecord] Error retrieving the record[%s] reason[%s]" % (name, err_msg))
    except Exception as e:
        logger.error("F[UpdateIpamRecord] Error retrieving the record[%s] reason[%s]" % (name, str(e)))
        raise CustomIpamGeneralException("F[UpdateIpamRecord] Error retrieving the record[%s] reason[%s]" % (name, str(e)))
    
    # Step 2: Raise a valid error message if the given record not found
    if not host_ref:
        err_msg = "F[UpdateIpamRecord] Record[%s] not found!" % name
        logger.error(err_msg)
        raise CustomIpamRecordNotFoundException(err_msg)
 
    # Step 3: Build payload data according to the new allocation_type
    # and call specific rest API and return new allocated items(alloc_info).
    rest_url = 'https://' + auth_params['server'] + '/wapi/' + \
                    auth_params['wapi_version'] + '/' + host_ref + '?_return_fields=ipv4addrs,ipv6addrs'
    v4_subnet = new_nw_and_subnet_list[0][1]
    v6_subnet = new_nw_and_subnet_list[0][2]
    if new_allocation_type == 'V4_ONLY' and old_allocation_type == 'V6_ONLY':
        nw_str_ip4 = '"ipv4addrs": [{"ipv4addr": "func:nextavailableip:' + v4_subnet +  ',' + network_view + '"}]'
        nw_str_ip6 = '"ipv6addrs":[]'
    elif new_allocation_type == 'V4_ONLY' and old_allocation_type == 'V4_V6':
        nw_str_ip4 = None
        nw_str_ip6 = '"ipv6addrs":[]'
    elif new_allocation_type == 'V6_ONLY' and old_allocation_type == 'V4_ONLY':
        nw_str_ip4 = '"ipv4addrs":[]'
        nw_str_ip6 = '"ipv6addrs": [{"ipv6addr": "func:nextavailableip:' + v6_subnet + ',' + network_view + '"}]'
    elif new_allocation_type == 'V6_ONLY' and old_allocation_type == 'V4_V6':
        nw_str_ip4 = '"ipv4addrs":[]'
        nw_str_ip6 = None
    elif new_allocation_type == 'V4_V6' and old_allocation_type == 'V4_ONLY':
        nw_str_ip4 = None
        nw_str_ip6 = '"ipv6addrs": [{"ipv6addr": "func:nextavailableip:' + v6_subnet + ',' + network_view + '"}]'
    elif new_allocation_type == 'V4_V6' and old_allocation_type == 'V6_ONLY':
        nw_str_ip4 = '"ipv4addrs": [{"ipv4addr": "func:nextavailableip:' + v4_subnet + ',' + network_view + '"}]'
        nw_str_ip6 = None
       
    payload = '{'
    if nw_str_ip4:
        payload += nw_str_ip4 
    if nw_str_ip4 and nw_str_ip6:
        payload += ','
    if nw_str_ip6: 
        payload += nw_str_ip6
    payload += '}'
    r = requests.put(url=rest_url, auth=(auth_params['username'], auth_params['password']),
                    verify=False, data=payload, timeout=30)
    logger.info("F[UpdateIpamRecord] req[%s %s] status_code[%s]" % (rest_url, payload, r.status_code))
    _check_and_raise_auth_error(r)
    if r.status_code not in [200, 201]:
        err_msg = "%d %s" % (r.status_code, r.text)
        logger.error("F[UpdateIpamRecord] Error updating the host record[%s] req[%s,%s] reason[%s]", name, rest_url, payload, err_msg)
        raise CustomIpamGeneralException("F[UpdateIpamRecord] Error updating the host record[%s] req[%s,%s] reason[%s]", 
            name, rest_url, payload, err_msg)
    r_json = r.json()
    alloc_info = {}
    if 'ipv4addrs' in r_json and r_json['ipv4addrs'][0]['ipv4addr']:
        alloc_info['v4_ip'] = r_json['ipv4addrs'][0]['ipv4addr']
        alloc_info['v4_subnet'] = v4_subnet
    if 'ipv6addrs' in r_json and r_json['ipv6addrs'][0]['ipv6addr']:
        alloc_info['v6_ip'] = r_json['ipv6addrs'][0]['ipv6addr']
        alloc_info['v6_subnet'] = v6_subnet
    return alloc_info 



############################################################################
# ========================================================================
# Copyright 2022 VMware, Inc.  All rights reserved. VMware Confidential
# ========================================================================
###

import importlib
import os
import argparse
import urllib3
import sys

REQUIRED_FUNCTIONS = ['TestLogin', 'GetAvailableNetworksAndSubnets', 'GetIpamRecord',
                      'CreateIpamRecord', 'DeleteIpamRecord', 'UpdateIpamRecord']

REQUIRED_CLASSES = ['CustomIpamAuthenticationErrorException', 'CustomIpamRecordNotFoundException',
                    'CustomIpamNoFreeIpException', 'CustomIpamNotImplementedException',
                    'CustomIpamGeneralException']

parser = argparse.ArgumentParser(description="This script tests a custom IPAM script")
parser.add_argument('script_path', help='File path of the custom IPAM script')
args = parser.parse_args()

# Import script module using its path
mod_local_path = args.script_path
dir_name = os.path.dirname(mod_local_path)
sys.path.insert(0, dir_name)

mod_name = os.path.basename(mod_local_path).replace(".py", "")

script = importlib.import_module(mod_name)


def check_required_functions_and_classes(script_object):
    object_list = dir(script_object)
    missing_functions_list, missing_classes_list = [], []

    for required_function in REQUIRED_FUNCTIONS:
        if required_function not in object_list:
            missing_functions_list.append(required_function)

    for required_class in REQUIRED_CLASSES:
        if required_class not in object_list:
            missing_classes_list.append(required_class)

    if missing_functions_list:
        missing_functions = ', '.join(missing_functions_list)
        assert 0, "Required functions: " + missing_functions + " not present in the script"

    if missing_classes_list:
        missing_classes = ', '.join(missing_classes_list)
        assert 0, "Required classes: " + missing_classes + " not present in the script"


def get_nw_and_subnet_value(auth_params, allocation_type):
    """
    In case the user does not enter an input for nw_and_subnet_list (record_info),
    use the available GetAvailableNetworksAndSubnets function to fetch a value for nw_and_subnet_list .
    nw_and_subnet_list is required for UpdateIpamRecord.
    """
    nw_and_subnet = script.GetAvailableNetworksAndSubnets(auth_params, allocation_type)[0]
    nw_and_subnet = [tuple(i for i in nw_and_subnet.values())]
    return nw_and_subnet


def get_auth_params_from_user():
    auth_params = {}

    i = int(input("Enter the number of fields for auth_params: "))

    while i > 0:
        try:
            key, value = input("Enter field name and value (example input: username user1): ").split()
        except:
            print("Error! Enter input in the correct format")
            continue
        auth_params[key] = value
        i -= 1

    return auth_params


def get_record_params_from_user():
    record_id = input("Enter record id: ")
    fqdns = input("Enter list of fqdns (example input: fqdns1 fqdns2 fqdns3): ").split()
    allocation_type = input("Enter allocation type ('V4_ONLY', 'V6_ONLY ', or 'V4_V6'): ")
    preferred_ip = input("Enter preferred ip (*optional: press Enter to skip): ")
    preferred_ip6 = input("Enter preferred ip6 (*optional: press Enter to skip):: ")

    print("\n------Enter network and subnet list (*optional)-----\n")

    nw_and_subnet_list = []
    flag = input("Do you want to enter a list? (y/n): ").lower()
    while flag == 'y':
        required_input = input("Enter network, v4_subnet, v6_subnet\n"
                               "--Example: nw,v4,v6\n"
                               "--Example (no v4): nw,,v6\n"
                               "--Example (no v4 v6): nw,,\n"
                               "Enter the value: ")
        if len(required_input):
            required_tuple = tuple(i.strip() for i in required_input.split(","))
            nw_and_subnet_list.append(required_tuple)
            flag = input("Do you want to enter another? (y/n): ").lower()
        else:
            break

    record_info = {"id": record_id,
                   "fqdns": fqdns,
                   "allocation_type": allocation_type,
                   "preferred_ip": preferred_ip,
                   "preferred_ip6": preferred_ip6,
                   "nw_and_subnet_list": nw_and_subnet_list
                   }

    return record_info


def login(auth_params):
    try:
        check_login = script.TestLogin(auth_params)
        print("\n******TestLogin PASSED******")
        print("Function output: ", check_login)
    except Exception as e:
        print("\n******TestLogin FAILED******")
        print(str(e))


def create_ipam_record(auth_params, record_info):
    try:
        check_create = script.CreateIpamRecord(auth_params, record_info)
        check_get = check_get_ipam_record(auth_params, record_info)

        # If GetIpamRecord is not implemented in the script
        if not check_get[0] and (type(check_get[1]) == script.CustomIpamNotImplementedException):
            print("Warning: GetIpamRecord not implemented, skipping additional check")
        else:
            assert check_get[0], str(check_get[1])  # Check if get returns True; print error in case of assertion error
            assert check_get[1] == check_create, "Error: GetIpamRecord didn't return the created record"

        print("\n******CreateIpamRecord PASSED******")
        print("Function output: ", check_create)
    except Exception as e:
        print("\n****** CreateIpamRecord FAILED******")
        print(str(e))


def check_get_ipam_record(auth_params, record_info):
    try:
        check_get_ipam_record = script.GetIpamRecord(auth_params, record_info)
        assert type(check_get_ipam_record) == dict, "Error: GetIpamRecord doesn't return a dictionary"
        return True, check_get_ipam_record
    except Exception as e:
        return False, e


def get_ipam_record(auth_params, record_info):
    check_get = check_get_ipam_record(auth_params, record_info)
    if check_get[0]:
        print("\n******GetIpamRecord PASSED******")
    else:
        print("\n******GetIpamRecord FAILED******")
        print(str(check_get[1]))


def get_available_networks_and_subnets(auth_params, ip_type):
    try:
        check_get_available_networks = script.GetAvailableNetworksAndSubnets(auth_params, ip_type)
        assert type(check_get_available_networks) == list, "\n******GetAvailableNetworksAndSubnets FAILED******"
        print("\n******GetAvailableNetworksAndSubnets PASSED******")
        print("Function output:", check_get_available_networks)
    except Exception as e:
        print("\n******GetAvailableNetworksAndSubnets FAILED******")
        print(str(e))


def update_ipam_record_menu(auth_params, record_info):
    record_info_new = record_info.copy()
    print("\n-----Update record-----")
    print("*****Editing*****")

    while True:
        choice = input(
            "Enter the field to update:\n1. id\n2. fqdns\n3. allocation_type\n4. preferred_ip\n5. preferred_ip6\n6. nw_and_subnet_list\n0. Finish Editing\n")
        if choice == '0':
            break
        elif choice == '1':
            record_info_new['id'] = input("Enter new id: ")
        elif choice == '2':
            record_info_new['fqdns'] = input("Enter new fqdns: ")
        elif choice == '3':
            record_info_new['allocation_type'] = input("Enter new allocation type: ")
        elif choice == '4':
            record_info_new['preferred_ip'] = input("Enter new preferred_ip: ")
        elif choice == '5':
            record_info_new['preferred_ip6'] = input("Enter new preferred_ip6: ")
        elif choice == '6':
            record_info_new['nw_and_subnet_list'] = input("Enter new nw_and_subnet_list: ")
        print("Value updated")

    update_ipam_record(auth_params, record_info_new, record_info)


def update_ipam_record(auth_params, new_record, old_record):
    """
    1. Run UpdateIpamRecord from the script
    2. Check if UpdateIpamRecord returns an output based on entered (updated) user input
    3. Check if GetIpamRecord returns the same output
    """

    try:
        # If no value provided for nw_and_subnet_list, get value from available networks
        if not len(old_record['nw_and_subnet_list']):
            old_record['nw_and_subnet_list'] = get_nw_and_subnet_value(auth_params, old_record['allocation_type'])

        if not len(new_record['nw_and_subnet_list']):
            new_record['nw_and_subnet_list'] = get_nw_and_subnet_value(auth_params, new_record['allocation_type'])

        check_update = script.UpdateIpamRecord(auth_params, new_record, old_record)
        assert type(check_update) == dict, "Error: UpdateIpamRecord doesn't return a dictionary"
        new_allocation_type = new_record['allocation_type']

        if new_allocation_type == 'V4_ONLY':
            assert 'v4_ip' in check_update, "Error: No v4_ip in updated record"
        elif new_allocation_type == 'V6_ONLY':
            assert 'v6_ip' in check_update, "Error:  No v6_ip in updated record"
        elif new_allocation_type == 'V4_V6':
            assert 'v4_ip' in check_update, "Error: No v4_ip in updated record"
            assert 'v6_ip' in check_update, "Error: No v6_ip in updated record"

        check_get = check_get_ipam_record(auth_params, new_record)

        # If GetIpamRecord is not implemented in the script
        if not check_get[0] and (type(check_get[1]) == script.CustomIpamNotImplementedException):
            print("Warning: GetIpamRecord not implemented, skipping additional check")
        else:
            assert check_get[0], str(check_get[1])  # Check if get returns True; print error in case of assertion error
            assert check_get[1] == check_update, "Error: GetIpamRecord didn't return the updated record"

        print("\n******UpdateIpamRecord PASSED******")
        print("Function output: ", check_update)
    except Exception as e:
        print("\n******UpdateIpamRecord FAILED******")
        print(str(e))


def delete_ipam_record(auth_params, record_info):
    try:
        check_delete = script.DeleteIpamRecord(auth_params, record_info)
        check_get = check_get_ipam_record(auth_params, record_info)
        assert not check_get[0], "Error: GetIpamRecord still returns the record. It did not get deleted"  # If delete succeeds, Get should return False
        print("\n******DeleteIpamRecord PASSED******")
        print("Function output:", check_delete)
    except Exception as e:
        print("\n******DeleteIpamRecord FAILED******")
        print(str(e))


def main():
    urllib3.disable_warnings()

    check_required_functions_and_classes(script)
    auth_params = get_auth_params_from_user()

    while True:
        user_input = int(input(
            "\n0. Exit\n1. Test Login\n2. Test GetIpamRecord\n3. Test GetAvailableNetworksAndSubnets\n4. Test CreateIpamRecord\n5. Test UpdateIpamRecord\n6. Test DeleteIpamRecord\n7. Test all\n\n"))

        if user_input == 0:
            break
        elif user_input == 1:
            print("------Testing TestLogin------")
            login(auth_params)

        elif user_input == 2:
            print("------Testing GetIpamRecord------")

            record_id = input("Enter record id: ")
            record_info = {"id": record_id}

            get_ipam_record(auth_params, record_info)

        elif user_input == 3:
            print("------Testing GetAvailableNetworksAndSubnets------")
            ip_type = input("Enter ip type ('V4_ONLY', 'V6_ONLY ', or 'V4_V6'): ")
            get_available_networks_and_subnets(auth_params, ip_type)

        elif user_input == 4:
            print("------Testing CreateIpamRecord------")
            print("------Enter record info------")
            record_info = get_record_params_from_user()
            create_ipam_record(auth_params, record_info)

        elif user_input == 5:
            print("------Testing UpdateIpamRecord------")
            print("------Enter old record info:------")
            record_info = get_record_params_from_user()

            update_ipam_record_menu(auth_params, record_info)

        elif user_input == 6:
            print("------Testing DeleteIpamRecord------")
            record_id = input("Enter record id: ")
            record_fqdns = input("Enter fqdns: ")

            record_info = {"id": record_id, "fqdns": record_fqdns}

            delete_ipam_record(auth_params, record_info)

        else:
            print("------Testing All------")
            record_info = {"id": "test123",
                           "allocation_type": "V4_ONLY",
                           "nw_and_subnet_list": []}

            # 1. Test login
            login(auth_params)

            # 2. Test GetIpamRecord
            get_ipam_record(auth_params, record_info)

            # 3. Test GetAvailableNetworksAndSubnets
            get_available_networks_and_subnets(auth_params, record_info['allocation_type'])

            # 4. Test CreateIpamRecord
            create_ipam_record(auth_params, record_info)

            # 5. Test UpdateIpamRecord (Convert allocation type from V4_ONLY to V6_ONLY)
            record_info_new = record_info.copy()
            record_info_new["allocation_type"] = 'V6_ONLY'

            update_ipam_record(auth_params, record_info_new, record_info)

            # 6. Test DeleteIpamRecord
            delete_ipam_record(auth_params, record_info)


if __name__ == '__main__':
    main()
